"""Define package errors."""


class VaillantError(Exception):
    """Define a base error."""

    pass


class RequestError(VaillantError):
    """Define an error related to invalid requests."""

    pass

class InvalidCredentialsError(VaillantError):
    """Define an error related to login issues.
       This error is only used internally.
    """
    pass

class InvalidAuthError(VaillantError):
    """Define an error related to login issues."""
    pass

class WebsocketServerClosedConnectionError(VaillantError):
    """Define an error when websocket server closed the connection."""
    pass

class WebsocketError(VaillantError):
    """Define an error related to generic websocket errors."""

    pass



"""Module to help store and retrieve credentials"""

from __future__ import annotations
from dataclasses import dataclass, field

import json
import base64

from .const import APP_ID, CONF_USERNAME, CONF_PASSWORD, CONF_UID, CONF_TOKEN, CONF_APP_ID, CONF_DID, CONF_MAC

@dataclass
class User:
    token: Token
    devices: list[Device] = field(default_factory=list)

@dataclass
class Device:
    id: str
    mac: str
    product_key: str
    product_name: str
    host: str
    ws_port: int
    wss_port: int
    wifi_soft_version: str
    wifi_hard_version: str
    mcu_soft_version: str
    mcu_hard_version: str
    is_online: bool
    model: str = ""
    serial_number: str = ""
    sno: str = ""

@dataclass
class Token:
    app_id: str
    username: str
    password: str
    token: str = ""
    uid: str = ""

    def serialize(self) -> str:
        """Serialize token object into a JSON string."""

        s = json.dumps({
            f"{CONF_APP_ID}": self.app_id,
            f"{CONF_USERNAME}": self.username,
            f"{CONF_PASSWORD}": self.password,
            f"{CONF_TOKEN}": self.token,
            f"{CONF_UID}": self.uid,
        })
        return base64.b64encode(s.encode("ascii")).decode("ascii")

    @classmethod
    def deserialize(cls, token: str) -> Token:
        """Deserialize JSON string into a token object."""

        data = json.loads(base64.b64decode(token.encode("ascii")))
        return Token(
            app_id=data[CONF_APP_ID],
            username=data[CONF_USERNAME],
            password=data[CONF_PASSWORD],
            token=data[CONF_TOKEN],
            uid=data[CONF_UID],
        )



"""Define an object to interact with the Websocket API."""
from __future__ import annotations

import asyncio
import logging
import json
from collections.abc import Awaitable, Callable
from typing import Any

import aiohttp

from .const import DEFAULT_API_VERSION, APP_ID, LOGGER, DEFAULT_USER_AGENT, STATE_CONNECTING, STATE_STOPPED, STATE_CONNECTED, STATE_SUBSCRIBED, STATE_DISCONNECTED, EVT_DEVICE_ATTR_UPDATE
from .errors import InvalidCredentialsError, InvalidAuthError, WebsocketServerClosedConnectionError
from .model import Token, Device

DEFAULT_WATCHDOG_TIMEOUT = 30


class WebsocketWatchdog:
    """Define a watchdog to kick the websocket connection at intervals."""

    def __init__(
        self,
        logger: logging.Logger,
        action: Callable[..., Awaitable],
        *,
        timeout_seconds: int = DEFAULT_WATCHDOG_TIMEOUT,
    ):
        """Initialize.

        Args:
            logger: The logger to use.
            action: The coroutine function to call when the watchdog expires.
            timeout_seconds: The number of seconds before the watchdog times out.
        """
        self._action = action
        self._logger = logger
        self._loop = asyncio.get_event_loop()
        self._timeout = timeout_seconds
        self._timer_task: asyncio.TimerHandle | None = None

    def cancel(self) -> None:
        """Cancel the watchdog."""
        if self._timer_task:
            self._timer_task.cancel()
            self._timer_task = None

    async def on_expire(self) -> None:
        """Log and act when the watchdog expires."""
        self._logger.info("Watchdog expired – calling %s",
                          self._action.__name__)
        await self._action()

    async def trigger(self) -> None:
        """Trigger the watchdog."""
        self._logger.info(
            "Watchdog triggered – sleeping for %s seconds", self._timeout)

        if self._timer_task:
            self._timer_task.cancel()

        self._timer_task = self._loop.call_later(
            self._timeout, lambda: asyncio.create_task(self.on_expire())
        )


class VaillantWebsocketClient:  # pylint: disable=too-many-instance-attributes
    """Define the websocket client."""

    def __init__(
        self,
        token: Token,
        device: Device,
        *,
        application_id: str = APP_ID,
        api_version: str = DEFAULT_API_VERSION,
        logger: logging.Logger = LOGGER,
        session: aiohttp.ClientSession | None = None,
        use_ssl: bool = True,
        verify_ssl: bool = True,
        max_retry_attemps: int | None = None,
        heartbeat_interval: int = DEFAULT_WATCHDOG_TIMEOUT,
    ) -> None:
        """Initialize.

        Args:
            application_key: An Ambient Weather application key.
            api_key: An Ambient Weather API key.
            api_version: The version of the API to query.
            logger: The logger to use.
        """

        self._session: aiohttp.ClientSession = session or aiohttp.ClientSession()
        self._use_ssl = use_ssl
        self._verify_ssl = verify_ssl
        self._token = token
        self._device = device
        self._application_id = application_id
        self._api_version = api_version
        self._async_on_subscribe_handler: Callable[[dict[str, Any]], Awaitable[None]] | None = None
        self._async_on_update_handler: Callable[..., Awaitable[None]] | None = None
        self._on_subscribe_handler: Callable[[dict[str, Any]], None] | None = None
        self._on_update_handler: Callable[..., None] | None = None
        self._logger = logger
        self._max_retry_attemps = max_retry_attemps
        self._state = None
        # self._ws_client = None
        self._watchdog = WebsocketWatchdog(logger, self.ping, timeout_seconds=heartbeat_interval)

        self._endpoint = f"ws://{self._device.host}:{self._device.ws_port}"
        if self._use_ssl:
            self._endpoint = f"wss://{self._device.host}:{self._device.wss_port}"

    @property
    def state(self):
        """Return the current state."""
        return self._state

    async def connect(self) -> None:
        self._state = STATE_CONNECTING

        try:
            async with self._session.ws_connect(
                f"{self._endpoint}/ws/app/v1",
                verify_ssl=self._verify_ssl,
                headers={
                    "User-Agent": DEFAULT_USER_AGENT,
                },
            ) as ws_client:
                self._ws_client = ws_client

                await ws_client.send_json({
                    "cmd": "login_req",
                    "data": {
                        "appid": self._token.app_id or self._application_id,
                        "uid": self._token.uid,
                        "token": self._token.token,
                        "p0_type": "attrs_v4",
                        "heartbeat_interval": 180,
                        "auto_subscribe": False,
                    },
                })

                login_ret = await ws_client.receive_json()
                if login_ret["cmd"] != "login_res" or login_ret["data"]["success"] is not True:
                    self._logger.error("login failed")
                    self._state = STATE_STOPPED
                    raise InvalidCredentialsError

                await ws_client.send_json({"cmd": "subscribe_req", "data": [{"did": self._device.id}]})
                await ws_client.send_json({"cmd": "c2s_read", "data": {"did": self._device.id}})

                self._state = STATE_CONNECTED

                await self._watchdog.trigger()

                async for message in ws_client:
                    if self.state == STATE_STOPPED:
                        break

                    if message.type == aiohttp.WSMsgType.TEXT:
                        msg: dict = message.json()
                        cmd = msg.get("cmd", "")
                        data = msg.get("data", {})

                        self._logger.debug("Recv: %s", message)

                        if cmd == "s2c_noti":
                            if data["did"] == self._device.id:
                                device_attrs = data["attrs"]
                                self._logger.debug(
                                    "Recv atrrs for device %s => %s", data["did"], device_attrs)
                                if self.state != STATE_SUBSCRIBED:
                                    if self._async_on_subscribe_handler is not None:
                                        await self._async_on_subscribe_handler(device_attrs)
                                    elif self._on_subscribe_handler is not None:
                                        self._on_subscribe_handler(device_attrs)
                                    self._state = STATE_SUBSCRIBED
                                
                                if self._async_on_update_handler is not None:
                                    await self._async_on_update_handler(EVT_DEVICE_ATTR_UPDATE, { "data": device_attrs })
                                elif self._on_update_handler is not None:
                                    self._on_update_handler(EVT_DEVICE_ATTR_UPDATE, { "data": device_attrs })
                        elif cmd == "pong":
                            await self._watchdog.trigger()
                        elif cmd == "s2c_invalid_msg":
                            self._logger.error("Server pushed an error msg: %s", message)
                            error_code = data["error_code"]
                            # {'cmd': 's2c_invalid_msg', 'data': {'error_code': 1009, 'msg': 'M2M socket has closed, please login again!'}}
                            # {'cmd': 's2c_invalid_msg', 'data': {'error_code': 1011, 'msg': 'No heartbeat!'}}
                            # TODO: other error code should treat more gracefully.
                            if error_code == 1009 or error_code == 1011:
                                raise WebsocketServerClosedConnectionError
                        else:
                            self._logger.info("Unhandled msg: %s", msg)

                    elif message.type == aiohttp.WSMsgType.CLOSED:
                        self._logger.warning(
                            "AIOHTTP websocket connection closed")
                        break

                    elif message.type == aiohttp.WSMsgType.ERROR:
                        self._logger.error("AIOHTTP websocket error")
                        break

        except InvalidCredentialsError as error:
            self._state = STATE_STOPPED
            raise InvalidAuthError
        except aiohttp.ClientResponseError as error:
            if error.code == 401:
                self._logger.error("Credentials rejected: %s", error)
                raise InvalidAuthError
            else:
                self._logger.error("Unexpected response received: %s", error)
            self._state = STATE_STOPPED
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError, WebsocketServerClosedConnectionError) as error:
            if self._max_retry_attemps is not None and self._failed_attempts >= self._max_retry_attemps:
                self._logger.error("Too many retry attemps. exit.")
                self._state = STATE_STOPPED
            elif self.state != STATE_STOPPED:
                retry_delay = min(2 ** (self._failed_attempts - 1) * 10, 300)
                self._failed_attempts += 1
                self._logger.error(
                    "Websocket connection failed, retrying in %ds: %s",
                    retry_delay,
                    error,
                )
                self._state = STATE_DISCONNECTED
                await asyncio.sleep(retry_delay)
        except Exception as error:  # pylint: disable=broad-except
            if self.state != STATE_STOPPED:
                self._logger.exception(
                    "Unexpected exception occurred: %s", error)
                self._state = STATE_STOPPED
        else:
            if self.state != STATE_STOPPED:
                self._state = STATE_DISCONNECTED
                await asyncio.sleep(5)

    async def listen(self):
        """Start the listening websocket."""
        self._failed_attempts = 0
        while self.state != STATE_STOPPED:
            await self.connect()

    async def close(self):
        """Close the listening websocket."""
        self._state = STATE_STOPPED
        self._watchdog.cancel()
        try:
            await self._ws_client.close()
        except Exception as error:
            self._logger.warning("Close websocket error: %s", error)

    async def ping(self) -> None:
        """Send ping to server."""
        await self.send_command("ping")

    async def send_command(self, cmd: str, data: dict[str, Any] | None = None) -> None:
        if self.state != STATE_STOPPED and self._ws_client is not None:
            msg: dict[str, Any] = {
                "cmd": cmd
            }
            if data is not None:
                msg["data"] = data

            await self._ws_client.send_json(msg)

    def async_on_subscribe(self, handler: Callable[[dict[str, Any]], Awaitable[None]]) -> None:
        self._async_on_subscribe_handler = handler

    def async_on_update(self, handler: Callable[..., Awaitable[None]]) -> None:
        self._async_on_update_handler = handler

    def on_subscribe(self, handler: Callable[[dict[str, Any]], None]) -> None:
        self._on_subscribe_handler = handler

    def on_update(self, handler: Callable[..., None]) -> None:
        self._on_update_handler = handler


"""Define package constants."""
import logging
import base64

LOGGER = logging.getLogger(__package__)

DEFAULT_API_VERSION = "1.3"

HOST_APP = "https://appapi.vaillant.com.cn"
HOST_API = "https://api.vaillant.com.cn"

APP_ID = str(base64.b64decode("MWY1YjkzNzNmZTk2NGYyNTgxYjBhM2EzZTI2MzYxZTY="))
DEFAULT_USER_AGENT = "Mozilla/5.0 (Linux; Android 7.1.2; Redmi 4X Build/N2G47H; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/66.0.3359.158 Mobile Safari/537.36"

STATE_CONNECTING = "connecting"
STATE_CONNECTED = "connected"
STATE_SUBSCRIBED = "subscribed"
STATE_STOPPED = "stopped"
STATE_DISCONNECTED = "disconnected"

EVT_DEVICE_ATTR_UPDATE = "device:attr:update"

CONF_USERNAME = "username"
CONF_PASSWORD = "password"
CONF_TOKEN = "token"
CONF_DID = "did"
CONF_UID = "uid"
CONF_HOST = "host"
CONF_PORT = "port"
CONF_MAC = "mac"
CONF_PRODUCT_NAME = "product_name"
CONF_APP_ID = "app_id"


from .api import VaillantApiClient
from .websocket import VaillantWebsocketClient
from .model import (
  Token,
  Device,
)
from .errors import (
  VaillantError,
  RequestError,
  InvalidAuthError
)
from .const import (
  EVT_DEVICE_ATTR_UPDATE,
  STATE_CONNECTING,
  STATE_CONNECTED,
  STATE_SUBSCRIBED,
  STATE_STOPPED,
  STATE_DISCONNECTED,
)

__all__ = [
  "VaillantApiClient",
  "VaillantWebsocketClient",
  "Token",
  "Device",
  "VaillantError",
  "RequestError",
  "InvalidAuthError",
  "EVT_DEVICE_ATTR_UPDATE",
  "STATE_CONNECTING",
  "STATE_CONNECTED",
  "STATE_SUBSCRIBED",
  "STATE_STOPPED",
  "STATE_DISCONNECTED",
]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vaillant_plus_cn_api']

package_data = \
{'': ['*']}

install_requires = \
['aiodns>=3.0.0,<4.0.0', 'aiohttp>=3.8.3,<4.0.0']

setup_kwargs = {
    'name': 'vaillant-plus-cn-api',
    'version': '0.2.0',
    'description': 'Python package for interacting with Vaillant devices sold in China mainland using API',
    'long_description': '# vaillant-plus-cn-api\nPython package for interacting with Vaillant devices sold in China mainland through APIs in Vaillant plus app.\n',
    'author': 'daxingplay',
    'author_email': 'daxingplay@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/daxingplay/vaillant-plus-cn-api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9',
}


setup(**setup_kwargs)



"""Define API client to interact with the Vaillant API."""
from __future__ import annotations

import asyncio
import logging
from datetime import date
from typing import Any, cast

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from .const import DEFAULT_API_VERSION, LOGGER, HOST_API, HOST_APP, APP_ID, DEFAULT_USER_AGENT
from .errors import RequestError, InvalidAuthError, RequestError
from .model import Token, Device

DEFAULT_LIMIT = 288
DEFAULT_TIMEOUT = 20


class VaillantApiClient:
    """Define the API client."""

    def __init__(
        self,
        *,
        application_id: str = APP_ID,
        user_agent: str = DEFAULT_USER_AGENT,
        api_version: str = DEFAULT_API_VERSION,
        logger: logging.Logger = LOGGER,
        session: ClientSession | None = None,
    ) -> None:
        """Initialize.

        Args:
            application_id: .
            api_version: The version of the API to query.
            logger: The logger to use.
            session: An optional aiohttp ClientSession.
        """
        self._application_id: str = application_id
        self._api_version: str = api_version
        self._user_agent: str = user_agent
        self._logger = logger
        self._session: ClientSession = session or self._new_session()

    def _new_session(self) -> ClientSession:
        return ClientSession(
            timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
            raise_for_status=False,
        )

    async def _request(
        self, method: str, url: str, **kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        """Make a request against the API.

        Args:
            method: An HTTP method.
            url: A relative API endpoint.
            **kwargs: Additional kwargs to send with the request.

        Returns:
            An API response payload.

        Raises:
            RequestError: Raised upon an underlying HTTP error.
        """

        kwargs.setdefault("params", {})
        kwargs.setdefault("headers", {})

        if use_running_session := self._session and not self._session.closed:
            session = self._session
        else:
            session = self._new_session()

        try:
            async with session.request(method, url, **kwargs) as resp:
                data = {}
                if 399 < resp.status and 500 > resp.status:
                    raise InvalidAuthError
                elif 200 == resp.status:
                    data = await resp.json()
                else:
                    resp.raise_for_status()
        except ClientError as err:
            raise RequestError(f"Error requesting data from {url}: {err}") from err
        finally:
            if not use_running_session:
                await session.close()

        self._logger.debug("Received data for %s: %s", url, data)

        return cast(dict[str, Any], data)

    async def login(self, username: str, password: str) -> Token:
        """Login to get uid and token."""
        data = {
            "appKey": self._application_id,
            "version": self._api_version,
            "data": {"username": username, "password": password},
        }
        headers = {
            "User-Agent": self._user_agent,
            "Referer": "http://localhost/main.html",
            "X-Requested-With": "com.vaillant.plus",
            "Origin": "http://localhost",
        }
        resp = await self._request(
            "post", f"{HOST_APP}/app/user/login", json=data, headers=headers)
        if resp is None or resp["code"] != "200" or resp["data"] is None:
            raise InvalidAuthError

        return Token(
            app_id=self._application_id,
            username=username,
            password=password,
            token=resp["data"]["token"],
            uid=resp["data"]["uid"],
        )

    async def get_device_list(self, token: str) -> list[Device]:
        """Get device list."""
        headers = {
            "User-Agent": "GizWifiSDK (v13.21121715)",
            "X-Gizwits-Application-Id": self._application_id,
            "X-Gizwits-User-token": token,
        }
        resp = await self._request(
          "get",
            f"{HOST_API}/app/bindings?show_disabled=0&limit=20&skip=0",
            headers=headers,
        )
        if resp.get("error_code") is not None:
            raise RequestError
        return [
            Device(
                id=d.get("did"),
                mac=d.get("mac"),
                product_key=d.get("product_key"),
                product_name=d.get("product_name"),
                host=d.get("host"),
                ws_port=d.get("ws_port"),
                wss_port=d.get("wss_port"),
                wifi_soft_version=d.get("wifi_soft_version", ""),
                wifi_hard_version=d.get("wifi_hard_version", ""),
                mcu_soft_version=d.get("mcu_soft_version", ""),
                mcu_hard_version=d.get("mcu_hard_version", ""),
                is_online=d.get("is_online"),
            )
            for d in resp["devices"]
        ]

    async def get_device_info(self, token: str, mac_addr: str) -> dict[str, str]:
        """Get device info"""
        headers = {
            "Authorization": token,
            "Version": self._api_version,
            "User-Agent": self._user_agent,
            "Referer": "http://localhost/main.html",
            "X-Requested-With": "com.vaillant.plus",
            "Origin": "http://localhost",
        }
        upper_mac = str.upper(mac_addr)
        resp = await self._request(
          "get",
            f"{HOST_APP}/app/device/sn/status?mac={upper_mac}",
            headers=headers,
        )
        code = resp.get("code", 0)
        if code == "505":
            raise InvalidAuthError

        if code != "200" or resp["data"] is None:
            raise RequestError

        return {
            "sno": resp["data"]["sno"],
            "mac": resp["data"]["mac"],
            "device_id": resp["data"]["gizDid"],
            "serial_number": resp["data"]["serialNumber"],
            "model": resp["data"]["model"],
            "status_code": resp["data"]["status"],
        }
